open Common;;
open Term;;
open Equations;;
open Typing_tree;;
open Substitutions;;
open Inference;;

let print_constraints s =
 print_string "\nConstraints:\n";
 let compt = ref 1 in
 let foo eq = Printf.printf "%s[%2d]  " (if decomposible eq then "D" else " ") !compt; incr compt; print_eq eq in
 List.iter foo s.eqs; print_newline ();;

let print_first_typed_term tm =
 print_string "\nTree:\n"; print_typed_term tm; print_newline ();;

let print_system_subst s =
 print_string "\nSubstitutions:\n";
 List.iter (fun sbs -> print_subst2 sbs; print_newline ()) (List.rev s.subst);;

let copy_system s =
 { eqs = s.eqs; subst = s.subst; ttree = s.ttree };;

let maximal_rank = ref 10;;

let check_rank print s =
 let rank = rank_typing_tree s.ttree in
 if print then Printf.printf "Current proof rank: %d\n" rank;
 if rank > !maximal_rank
 then (Printf.printf "\nMaximal rank exceeded !\nThis term is not typable with rank <= %d.\n" !maximal_rank; exit 1);;

let possible s =
 let l = ref [] in
 for i = List.length s.eqs - 1 downto 0 do
  if decomposible (List.nth s.eqs i) then l := (i+1)::!l
 done;
 !l;;

let all_proofs s =
 let compt = ref 0 in
 let rec loop s =
  check_rank false s;
  let l = possible s in
  if l = []
  then
   begin
    try
     while true do check_rank false s; step s None done
    with Next -> final s;
                 check_rank false s;
                 print_system_subst s; print_newline ();
                 proof s;
                 incr compt;
                 print_newline ()
   end
  else
   for i = 0 to List.length l - 1 do
    let s' = copy_system s in
    try
     step s' (Some (List.nth l i));
     loop s'
    with
     e -> print_int (List.nth l i); print_string " "; raise e
   done
 in
 loop s;
 print_string "Number of proofs tested: "; print_int !compt; print_newline ();
 exit 0;;

let interface_loop s =
 try
  while true do   (* escape via Next (or bug) *)
   print_constraints s;
   check_rank true s;
   print_string "\n[S]tep(n)   [G]o   [A]ll: "; flush stdout;
   let line = read_line () in
   if String.length line > 0 then
   begin
    let c = String.get line 0 in
    print_newline ();
    match c with
       's' | 'S' -> let numopt = try Some (int_of_string (String.sub line 1 (String.length line - 1)))
                                 with Failure "int_of_string" -> None in
                    step s numopt
     | 'g' | 'G' -> while true do check_rank false s; step s None done   (* escape via Next (or bug) *)
     | 'a' | 'A' -> all_proofs s     (* never come back *)
     | _ -> Printf.printf "Unknown command: %s\n" line
   end
  done
 with
  Next -> final s; print_newline ();
          check_rank true s;
          print_system_subst s; print_newline ();
          proof s;;

let main _ =
 try
  print_string "     Intersection Types Inference System\n\n";
  if Array.length Sys.argv > 2
  || Array.length Sys.argv = 2 && (Sys.argv.(1) = "-h" || Sys.argv.(1) = "-help" || Sys.argv.(1) = "--help")
  then raise Exit
  else if Array.length Sys.argv = 2
  then (maximal_rank := (try int_of_string Sys.argv.(1) with _ -> raise Exit);
        Printf.printf "Maximal rank: %d\n" !maximal_rank)
  else Printf.printf "Using default maximal rank: %d\n" !maximal_rank;
  print_string "\nTerm: "; flush stdout;
  let lexbuf = Lexing.from_channel stdin in
  let term = try Parser.instr Lexer.token lexbuf
             with Parsing.Parse_error -> runtime_msg "Syntax error" in
  let tm = term_to_typed_term term in
  print_first_typed_term tm;
  interface_loop { eqs = build_constr tm; subst = []; ttree = typed_term_to_typing_tree tm }
 with
    Lexer.Eof | End_of_file -> print_newline ()
  | Runtime s -> Printf.printf "\nUser error: %s\n" s
  | Failure s -> Printf.printf "\nInternal error: %s\nPlease send me the details of this run !\n" s
  | Exit -> Printf.printf "Usage: infer [max_rank]\n  max_rank: maximal rank (%d by default)\n" !maximal_rank;;

let _ = main ();;
